/**
 *
 * @copyright &copy; 2010 - 2021, Fraunhofer-Gesellschaft zur Foerderung der angewandten Forschung e.V.
 * All rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * We kindly request you to use one or more of the following phrases to refer to
 * foxBMS in your hardware, software, documentation or advertising materials:
 *
 * - &Prime;This product uses parts of foxBMS&reg;&Prime;
 * - &Prime;This product includes parts of foxBMS&reg;&Prime;
 * - &Prime;This product is derived from foxBMS&reg;&Prime;
 *
 */

/**
 * @file    diag_cfg.c
 * @author  foxBMS Team
 * @date    2019-11-28 (date of creation)
 * @updated 2021-09-29 (date of last update)
 * @ingroup ENGINE_CONFIGURATION
 * @prefix  DIAG
 *
 * @brief   Diagnostic module configuration
 *
 * The configuration of the different diagnosis events defined in diag_cfg.h is set in the array
 * DIAG_ID_cfg[], e.g. initialization errors or runtime errors.
 *
 * Every entry of the DIAG_ID_cfg[] array consists of
 *  - name of the diagnosis event (defined in diag_cfg.h)
 *  - type of diagnosis event
 *  - diagnosis sensitivity (after how many occurrences event is counted as error)
 *  - enabling of the recording for diagnosis event
 *  - enabling of the diagnosis event
 *  - callback function for diagnosis event if wished, otherwise DIAG_dummy_callback
 *
 * The system monitoring configuration defined in diag_cfg.h is set in the array
 * sysm_ch_cfg[]. The system monitoring is at the moment only used for
 * supervising the cyclic/periodic tasks.
 *
 * Every entry of the sysm_ch_cfg[] consists of
 *  - enum of monitored object
 *  - type of monitored object (at the moment only DIAG_SYSMON_CYCLICTASK is supported)
 *  - maximum delay in [ms] in which the object needs to call the SYSM_Notify function defined in diag.c
 *  - enabling of the recording for system monitoring
 *  - enabling of the system monitoring for the monitored object
 *  - callback function if system monitoring notices an error if wished, otherwise SYSM_DummyCallback
 */

/*========== Includes =======================================================*/
#include "general.h"

#include "diag_cfg.h"

#include "database.h"
#include "diag_cbs.h"

/*========== Macros and Definitions =========================================*/
/** value of #DIAG_ID_MAX (as a define for the pre-processor) */
#define DIAG_ID_MAX_FOR_INIT (73u)

static_assert(DIAG_ID_MAX_FOR_INIT == (uint16_t)DIAG_ID_MAX, "Both values need to be identical.");

/*========== Static Constant and Variable Definitions =======================*/
/** local copy of the #DATA_BLOCK_ERRORSTATE_s table */
static DATA_BLOCK_ERRORSTATE_s diag_tableErrorFlags = {.header.uniqueId = DATA_BLOCK_ID_ERRORSTATE};

/** local copy of the #DATA_BLOCK_MOL_FLAG_s table */
static DATA_BLOCK_MOL_FLAG_s diag_tableMolFlags = {.header.uniqueId = DATA_BLOCK_ID_MOL_FLAG};

/** local copy of the #DATA_BLOCK_RSL_FLAG_s table */
static DATA_BLOCK_RSL_FLAG_s diag_tableRslFlags = {.header.uniqueId = DATA_BLOCK_ID_RSL_FLAG};

/** local copy of the #DATA_BLOCK_MSL_FLAG_s table */
static DATA_BLOCK_MSL_FLAG_s diag_tableMslFlags = {.header.uniqueId = DATA_BLOCK_ID_MSL_FLAG};

const DIAG_DATABASE_SHIM_s diag_kDatabaseShim = {
    .pTableError = &diag_tableErrorFlags,
    .pTableMol   = &diag_tableMolFlags,
    .pTableRsl   = &diag_tableRslFlags,
    .pTableMsl   = &diag_tableMslFlags,
};

/*========== Static Function Prototypes =====================================*/

/*========== Extern Constant and Variable Definitions =======================*/
/** variable tracking the state of the diag channels */
DIAG_ID_CFG_s DIAG_ID_cfg[] = {
    /* clang-format off */
    {DIAG_ID_FLASHCHECKSUM, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_NO_DELAY, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_DummyCallback},
    {DIAG_ID_SYSTEMMONITORING, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_NO_DELAY, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorSystemMonitoring},
    {DIAG_ID_CONFIGASSERT, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_NO_DELAY, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_DummyCallback},

    {DIAG_ID_INTERLOCK_FEEDBACK, DIAG_ERROR_INTERLOCK_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_INTERLOCK_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorInterlock},

    {DIAG_ID_CELLVOLTAGE_OVERVOLTAGE_MSL, DIAG_ERROR_VOLTAGE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERVOLTAGE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvervoltage},
    {DIAG_ID_CELLVOLTAGE_OVERVOLTAGE_RSL, DIAG_ERROR_VOLTAGE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvervoltage},
    {DIAG_ID_CELLVOLTAGE_OVERVOLTAGE_MOL, DIAG_ERROR_VOLTAGE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvervoltage},

    {DIAG_ID_CELLVOLTAGE_UNDERVOLTAGE_MSL, DIAG_ERROR_VOLTAGE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_UNDERVOLTAGE_ms,  DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndervoltage},
    {DIAG_ID_CELLVOLTAGE_UNDERVOLTAGE_RSL, DIAG_ERROR_VOLTAGE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndervoltage},
    {DIAG_ID_CELLVOLTAGE_UNDERVOLTAGE_MOL, DIAG_ERROR_VOLTAGE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndervoltage},

    {DIAG_ID_TEMP_OVERTEMPERATURE_CHARGE_MSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_TEMPERATURE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureCharge},
    {DIAG_ID_TEMP_OVERTEMPERATURE_CHARGE_RSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureCharge},
    {DIAG_ID_TEMP_OVERTEMPERATURE_CHARGE_MOL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureCharge},

    {DIAG_ID_TEMP_OVERTEMPERATURE_DISCHARGE_MSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_TEMPERATURE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureDischarge},
    {DIAG_ID_TEMP_OVERTEMPERATURE_DISCHARGE_RSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureDischarge},
    {DIAG_ID_TEMP_OVERTEMPERATURE_DISCHARGE_MOL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvertemperatureDischarge},

    {DIAG_ID_TEMP_UNDERTEMPERATURE_CHARGE_MSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_TEMPERATURE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureCharge},
    {DIAG_ID_TEMP_UNDERTEMPERATURE_CHARGE_RSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureCharge},
    {DIAG_ID_TEMP_UNDERTEMPERATURE_CHARGE_MOL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureCharge},

    {DIAG_ID_TEMP_UNDERTEMPERATURE_DISCHARGE_MSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_TEMPERATURE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureDischarge},
    {DIAG_ID_TEMP_UNDERTEMPERATURE_DISCHARGE_RSL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureDischarge},
    {DIAG_ID_TEMP_UNDERTEMPERATURE_DISCHARGE_MOL, DIAG_ERROR_TEMPERATURE_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorUndertemperatureDischarge},

    {DIAG_ID_OVERCURRENT_CHARGE_CELL_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},
    {DIAG_ID_OVERCURRENT_CHARGE_CELL_RSL, DIAG_ERROR_CURRENT_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},
    {DIAG_ID_OVERCURRENT_CHARGE_CELL_MOL, DIAG_ERROR_CURRENT_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},

    {DIAG_ID_OVERCURRENT_DISCHARGE_CELL_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},
    {DIAG_ID_OVERCURRENT_DISCHARGE_CELL_RSL, DIAG_ERROR_CURRENT_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},
    {DIAG_ID_OVERCURRENT_DISCHARGE_CELL_MOL, DIAG_ERROR_CURRENT_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},

    {DIAG_ID_STRING_OVERCURRENT_CHARGE_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},
    {DIAG_ID_STRING_OVERCURRENT_CHARGE_RSL, DIAG_ERROR_CURRENT_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},
    {DIAG_ID_STRING_OVERCURRENT_CHARGE_MOL, DIAG_ERROR_CURRENT_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},

    {DIAG_ID_STRING_OVERCURRENT_DISCHARGE_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},
    {DIAG_ID_STRING_OVERCURRENT_DISCHARGE_RSL, DIAG_ERROR_CURRENT_SENSITIVITY_RSL, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},
    {DIAG_ID_STRING_OVERCURRENT_DISCHARGE_MOL, DIAG_ERROR_CURRENT_SENSITIVITY_MOL, DIAG_INFO, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},

    {DIAG_ID_PACK_OVERCURRENT_DISCHARGE_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentDischarge},
    {DIAG_ID_PACK_OVERCURRENT_CHARGE_MSL, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorOvercurrentCharge},

    {DIAG_ID_CURRENT_ON_OPEN_STRING, DIAG_ERROR_CURRENT_SENSITIVITY_MSL, DIAG_FATAL_ERROR, DIAG_DELAY_OVERCURRENT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorCurrentOnOpenString},

    {DIAG_ID_LTC_SPI, DIAG_ERROR_LTC_SPI_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_AFE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorLtc},
    {DIAG_ID_LTC_PEC, DIAG_ERROR_LTC_PEC_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_AFE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorLtc},
    {DIAG_ID_LTC_MUX, DIAG_ERROR_LTC_MUX_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_AFE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorLtc},
    {DIAG_ID_LTC_CONFIG, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_DELAY_AFE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorLtc},
    {DIAG_ID_OPEN_WIRE, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorLtc},

    /* Communication events */
    {DIAG_ID_CAN_TIMING, DIAG_ERROR_CAN_TIMING_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_CAN_TIMING_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_TIMING, DIAG_ErrorCanTiming},
    {DIAG_ID_CAN_RX_QUEUE_FULL, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED,  DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorCanRxQueueFull},
    {DIAG_ID_CAN_CC_RESPONDING, DIAG_ERROR_CAN_TIMING_CC_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_EC_CC_TIMING_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorCurrentSensor},
    {DIAG_ID_CAN_EC_RESPONDING, DIAG_ERROR_CAN_TIMING_CC_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_EC_CC_TIMING_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorCurrentSensor},
    {DIAG_ID_CURRENT_SENSOR_RESPONDING, DIAG_ERROR_CAN_SENSOR_SENSITIVITY,  DIAG_FATAL_ERROR, DIAG_DELAY_CURRENT_SENSOR_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorCurrentSensor},
    {DIAG_ID_SBC_FIN_STATE, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_DELAY_SBC_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_Sbc},
    {DIAG_ID_SBC_RSTB_STATE, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_FATAL_ERROR, DIAG_DELAY_SBC_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_Sbc},
    {DIAG_ID_PLAUSIBILITY_PACK_VOLTAGE, DIAG_ERROR_SENSITIVITY_TENTH_EVENT, DIAG_FATAL_ERROR, DIAG_DELAY_PL_PACK_VOLTAGE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorPlausibility},
    {DIAG_ID_STRING_CONTACTOR_FEEDBACK, DIAG_ERROR_CONTACTOR_FEEDBACK_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_CONTACTOR_FEEDBACK_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_StringContactorFeedback},
    {DIAG_ID_PRECHARGE_CONTACTOR_FEEDBACK, DIAG_ERROR_CONTACTOR_FEEDBACK_SENSITIVITY, DIAG_FATAL_ERROR, DIAG_DELAY_CONTACTOR_FEEDBACK_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_PrechargeContactorFeedback},
    {DIAG_ID_PLAUSIBILITY_CELL_VOLTAGE, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_PlausibilityCheck},
    {DIAG_ID_PLAUSIBILITY_CELL_VOLTAGE_SPREAD, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_PlausibilityCheck},
    {DIAG_ID_PLAUSIBILITY_CELL_TEMP, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_PlausibilityCheck},
    {DIAG_ID_PLAUSIBILITY_CELL_TEMPERATURE_SPREAD, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_PlausibilityCheck},
    {DIAG_ID_AFE_CELL_VOLTAGE_MEAS_ERROR, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},
    {DIAG_ID_AFE_CELL_TEMPERATURE_MEAS_ERROR, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},
    {DIAG_ID_BASE_CELL_VOLTAGE_MESUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},
    {DIAG_ID_REDUNDANCY0_CELL_VOLTAGE_MESUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},
    {DIAG_ID_BASE_CELL_TEMPERATURE_MESUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},
    {DIAG_ID_REDUNDANCY0_CELL_TEMPERATURE_MESUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_FIRST_EVENT, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorAfe},

    {DIAG_ID_DEEP_DISCHARGE_DETECTED, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_DEEP_DISCHARGE_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorDeepDischarge},

    {DIAG_ID_CURRENT_MEASUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_TIMEOUT_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorCurrentMeasurement},
    {DIAG_ID_CURRENT_MEASUREMENT_ERROR, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_ERROR_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorCurrentMeasurement},

    {DIAG_ID_CURRENT_SENSOR_V1_MEASUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_TIMEOUT_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorHighVoltageMeasurement},
    {DIAG_ID_CURRENT_SENSOR_V3_MEASUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_TIMEOUT_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorHighVoltageMeasurement},
    {DIAG_ID_CURRENT_SENSOR_POWER_MEASUREMENT_TIMEOUT, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_TIMEOUT_ms, DIAG_RECORDING_ENABLED, DIAG_CAN_SENSOR_PRESENT, DIAG_ErrorPowerMeasurement},
    {DIAG_ID_POWER_MEASUREMENT_ERROR, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_FATAL_ERROR, DIAG_DELAY_REDUNDANCY_MEAS_ERROR_ms, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_ErrorPowerMeasurement},

    {DIAG_ID_INSULATION_MEASUREMENT_INVALID, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_Insulation},
    {DIAG_ID_INSULATION_ERROR, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_Insulation},
    {DIAG_ID_INSULATION_GROUND_ERROR, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_WARNING, DIAG_DELAY_DISCARDED, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_Insulation},

    {DIAG_ID_I2C_PEX_ERROR, DIAG_ERROR_SENSITIVITY_HIGH, DIAG_INFO, DIAG_NO_DELAY, DIAG_RECORDING_ENABLED, DIAG_EVALUATION_ENABLED, DIAG_I2cPex},
    /* clang-format on */
};

DIAG_DEV_s diag_device = {
    .nr_of_ch             = sizeof(DIAG_ID_cfg) / sizeof(DIAG_ID_CFG_s),
    .ch_cfg               = &DIAG_ID_cfg[0],
    .numberOfFatalErrors  = 0u,
    .pFatalErrorLinkTable = {REPEAT_U(NULL_PTR, STRIP(DIAG_ID_MAX_FOR_INIT))},
};

/*========== Static Function Implementations ================================*/

/*========== Extern Function Implementations ================================*/
void DIAG_UpdateFlags(void) {
    DATA_WRITE_DATA(&diag_tableErrorFlags, &diag_tableMolFlags, &diag_tableRslFlags, &diag_tableMslFlags);
}

/*========== Externalized Static Function Implementations (Unit Test) =======*/
